import * as ts from 'typescript';
import { MANTASTYLE_RUNTIME_NAME } from './constants';
import { isOptionalTypeNode, isRestTypeNode } from './typescript';
/*
type X = {
  haha: string
}

type X = {
  'haha': string
}

type X = {
  [key: string]: string
}

type X = {
  [key in Y]: string
}
*/
export function createConstVariableStatement(name, initializer) {
    return ts.createVariableStatement(undefined, ts.createVariableDeclarationList([ts.createVariableDeclaration(name, undefined, initializer)], ts.NodeFlags.Const));
}
export function createTypeAliasDeclaration(node) {
    var name = node.name.getText();
    var typeParameters = node.typeParameters || ts.createNodeArray();
    var jsdocTags = ts.getJSDocTags(node);
    var varCreation = createConstVariableStatement(name, ts.createFunctionExpression(undefined, undefined, undefined, undefined, [], undefined, ts.createBlock([
        ts.createReturn(createRuntimeFunctionCall('TypeAliasDeclaration', [
            ts.createStringLiteral(name),
            ts.createArrowFunction(undefined, undefined, [
                ts.createParameter(undefined, undefined, undefined, 'typeFactory', undefined, undefined, undefined),
            ], undefined, undefined, ts.createBlock(createTypeParameters(typeParameters, typeParameters, 'typeFactory').concat([
                createConstVariableStatement('type', createMantaStyleRuntimeObject(node.type, typeParameters)),
                ts.createReturn(ts.createIdentifier('type')),
            ]), true)),
            generateJSDocParam(jsdocTags),
        ])),
    ])));
    var registerToRuntime = createRuntimeFunctionCall('registerType', [
        ts.createStringLiteral(name),
        ts.createIdentifier(name),
    ]);
    varCreation.modifiers = node.modifiers;
    return [varCreation, registerToRuntime];
}
export function createRuntimeFunctionCall(methodName, argArray, variableName) {
    if (variableName === void 0) { variableName = MANTASTYLE_RUNTIME_NAME; }
    return ts.createCall(ts.createPropertyAccess(ts.createIdentifier(variableName), ts.createIdentifier(methodName)), undefined, argArray);
}
function createRuntimePropertyRef(propertyName, variable) {
    if (variable === void 0) { variable = MANTASTYLE_RUNTIME_NAME; }
    return ts.createPropertyAccess(ts.createIdentifier(variable), ts.createIdentifier(propertyName));
}
function createTypeParameters(typeParameters, referenceTypeParameters, caller) {
    var statements = [];
    if (typeParameters) {
        for (var _i = 0, typeParameters_1 = typeParameters; _i < typeParameters_1.length; _i++) {
            var param = typeParameters_1[_i];
            statements.push(createConstVariableStatement(param.name.getText(), createRuntimeFunctionCall('TypeParameter', [
                ts.createStringLiteral(param.name.getText())
            ].concat((param.constraint
                ? [
                    createMantaStyleRuntimeObject(param.constraint, referenceTypeParameters),
                ]
                : []), (param.default
                ? [
                    createMantaStyleRuntimeObject(param.default, referenceTypeParameters),
                ]
                : [])), caller)));
        }
    }
    return statements;
}
function createPropertyName(node) {
    var name = node.name;
    if (ts.isStringLiteral(name)) {
        return name;
    }
    else if (ts.isIdentifier(name)) {
        return ts.createStringLiteral(name.getText());
    }
    throw new Error('Unsupported types when creating property name.');
}
function isGenericTypeReference(typeName, typeParameters) {
    return ((typeParameters &&
        typeParameters.some(function (item) { return item.name.getText() === typeName; })) ||
        false);
}
function createTypeReferenceOrIdentifier(node, typeParameters) {
    var nodeName = node.getText();
    if (isGenericTypeReference(nodeName, typeParameters)) {
        return ts.createIdentifier(nodeName);
    }
    return createMantaStyleRuntimeObject(node, typeParameters);
}
function generateJSDocParam(jsdocArray) {
    return ts.createArrayLiteral(jsdocArray.map(function (tag) {
        return ts.createObjectLiteral([
            ts.createPropertyAssignment('key', ts.createStringLiteral(tag.tagName.text)),
            ts.createPropertyAssignment('value', ts.createStringLiteral(tag.comment || '')),
        ]);
    }));
}
function createTypeLiteralProperties(members, typeParameters) {
    var statements = [];
    for (var _i = 0, members_1 = members; _i < members_1.length; _i++) {
        var member = members_1[_i];
        var jsdocArray = ts.getJSDocTags(member);
        if (ts.isPropertySignature(member) && member.type) {
            statements.push(ts.createExpressionStatement(createRuntimeFunctionCall('property', [
                createPropertyName(member),
                createTypeReferenceOrIdentifier(member.type, typeParameters),
                member.questionToken ? ts.createTrue() : ts.createFalse(),
                generateJSDocParam(jsdocArray),
            ], 'typeLiteral')));
        }
        else if (ts.isIndexSignatureDeclaration(member)) {
            if (member.parameters.length === 1 && member.type) {
                var parameter = member.parameters[0];
                if (parameter.type) {
                    statements.push(ts.createExpressionStatement(createRuntimeFunctionCall('computedProperty', [
                        ts.createStringLiteral(parameter.name.getText()),
                        createTypeReferenceOrIdentifier(parameter.type, typeParameters),
                        createTypeReferenceOrIdentifier(member.type, typeParameters),
                        ts.createNumericLiteral('0'),
                        member.questionToken ? ts.createTrue() : ts.createFalse(),
                        generateJSDocParam(jsdocArray),
                    ], 'typeLiteral')));
                }
            }
        }
    }
    return statements;
}
function createBatchTypeFunctionAll(node, typeParameters, functionName) {
    return createRuntimeFunctionCall(functionName, [
        ts.createArrayLiteral(node.types.map(function (item) {
            return createTypeReferenceOrIdentifier(item, typeParameters);
        })),
    ]);
}
function createUnionType(node, typeParameters) {
    return createBatchTypeFunctionAll(node, typeParameters, 'UnionType');
}
function createIntersectionType(node, typeParameters) {
    return createBatchTypeFunctionAll(node, typeParameters, 'IntersectionType');
}
function createLiteralType(node) {
    return createRuntimeFunctionCall('Literal', [
        ts.createIdentifier(node.literal.getText()),
    ]);
}
function createArrayType(node, typeParameters) {
    return createRuntimeFunctionCall('ArrayType', [
        createMantaStyleRuntimeObject(node.elementType, typeParameters),
    ]);
}
function createMappedType(node, typeParameters) {
    var referenceTypeParameters = ts.createNodeArray(typeParameters.concat([
        node.typeParameter,
    ]));
    if (node.typeParameter.constraint && node.type) {
        return createRuntimeFunctionCall('MappedType', [
            ts.createArrowFunction(undefined, undefined, [
                ts.createParameter(undefined, undefined, undefined, 'mappedType', undefined, undefined, undefined),
            ], undefined, undefined, ts.createBlock(createTypeParameters(ts.createNodeArray([node.typeParameter]), referenceTypeParameters, 'mappedType').concat([
                ts.createExpressionStatement(createRuntimeFunctionCall('setQuestionToken', [
                    ts.createNumericLiteral(String(node.questionToken
                        ? node.questionToken.kind === ts.SyntaxKind.MinusToken
                            ? 2 /* MinusToken */
                            : 1 /* QuestionToken */
                        : 0 /* None */)),
                ], 'mappedType')),
                ts.createExpressionStatement(createRuntimeFunctionCall('setConstraint', [
                    createMantaStyleRuntimeObject(node.typeParameter.constraint, referenceTypeParameters),
                ], 'mappedType')),
                ts.createReturn(createMantaStyleRuntimeObject(node.type, referenceTypeParameters)),
            ]), true)),
        ]);
    }
    else {
        throw new Error('Invalid MappedType');
    }
}
export function createTypeLiteral(node, typeParameters) {
    return createRuntimeFunctionCall('TypeLiteral', [
        ts.createArrowFunction(undefined, undefined, [
            ts.createParameter(undefined, undefined, undefined, 'typeLiteral', undefined, undefined, undefined),
        ], undefined, undefined, ts.createBlock(createTypeLiteralProperties(node.members, typeParameters), true)),
    ]);
}
function createTypeReference(node, typeParameters) {
    var typeName = node.typeName.getText();
    // Magic type to convert query string to types
    if (typeName === 'unstable_Query' && node.typeArguments) {
        var queryKey = node.typeArguments[0];
        if (ts.isLiteralTypeNode(queryKey) &&
            ts.isStringLiteral(queryKey.literal)) {
            return createRuntimeFunctionCall('TypeReference', [
                ts.createStringLiteral("" + "@@URLQuery/" /* URLQuery */ + queryKey.literal.text),
            ]);
        }
        else {
            throw Error('key in unstable_Query is not a string literal.');
        }
    }
    var typeReferenceNode = isGenericTypeReference(typeName, typeParameters)
        ? ts.createIdentifier(typeName)
        : createRuntimeFunctionCall('TypeReference', [
            ts.createStringLiteral(typeName),
        ]);
    if (node.typeArguments) {
        return ts.createCall(ts.createPropertyAccess(typeReferenceNode, 'argumentTypes'), [], [
            ts.createArrayLiteral(node.typeArguments.map(function (type) {
                return createTypeReferenceOrIdentifier(type, typeParameters);
            })),
        ]);
    }
    else {
        return typeReferenceNode;
    }
}
export function createMantaStyleRuntimeObject(node, typeParameters) {
    if (ts.isTypeLiteralNode(node)) {
        return createTypeLiteral(node, typeParameters);
    }
    else if (ts.isUnionTypeNode(node)) {
        return createUnionType(node, typeParameters);
    }
    else if (ts.isIntersectionTypeNode(node)) {
        return createIntersectionType(node, typeParameters);
    }
    else if (ts.isLiteralTypeNode(node)) {
        return createLiteralType(node);
    }
    else if (ts.isParenthesizedTypeNode(node)) {
        return createRuntimeFunctionCall('ParenthesizedType', [
            createMantaStyleRuntimeObject(node.type, typeParameters),
        ]);
    }
    else if (ts.isArrayTypeNode(node)) {
        return createArrayType(node, typeParameters);
    }
    else if (ts.isTupleTypeNode(node)) {
        return createRuntimeFunctionCall('TupleType', [
            ts.createArrayLiteral(node.elementTypes.map(function (type) {
                return createMantaStyleRuntimeObject(type, typeParameters);
            })),
        ]);
    }
    else if (isOptionalTypeNode(node)) {
        return createRuntimeFunctionCall('OptionalType', [
            createMantaStyleRuntimeObject(node.type, typeParameters),
        ]);
    }
    else if (isRestTypeNode(node) && ts.isArrayTypeNode(node.type)) {
        return createRuntimeFunctionCall('RestType', [
            createMantaStyleRuntimeObject(node.type.elementType, typeParameters),
        ]);
    }
    else if (ts.isIndexedAccessTypeNode(node)) {
        return createRuntimeFunctionCall('IndexedAccessType', [
            createMantaStyleRuntimeObject(node.objectType, typeParameters),
            createMantaStyleRuntimeObject(node.indexType, typeParameters),
        ]);
    }
    else if (ts.isConditionalTypeNode(node)) {
        return createRuntimeFunctionCall('ConditionalType', [
            createMantaStyleRuntimeObject(node.checkType, typeParameters),
            createMantaStyleRuntimeObject(node.extendsType, typeParameters),
            createMantaStyleRuntimeObject(node.trueType, typeParameters),
            createMantaStyleRuntimeObject(node.falseType, typeParameters),
        ]);
    }
    else if (ts.isMappedTypeNode(node)) {
        return createMappedType(node, typeParameters);
    }
    else if (node.kind === ts.SyntaxKind.NumberKeyword) {
        return createRuntimePropertyRef('NumberKeyword');
    }
    else if (node.kind === ts.SyntaxKind.BooleanKeyword) {
        return createRuntimePropertyRef('BooleanKeyword');
    }
    else if (node.kind === ts.SyntaxKind.ObjectKeyword) {
        return createRuntimePropertyRef('ObjectKeyword');
    }
    else if (node.kind === ts.SyntaxKind.StringKeyword) {
        return createRuntimePropertyRef('StringKeyword');
    }
    else if (node.kind === ts.SyntaxKind.NeverKeyword) {
        return createRuntimePropertyRef('NeverKeyword');
    }
    else if (node.kind === ts.SyntaxKind.NullKeyword) {
        return createRuntimePropertyRef('NullKeyword');
    }
    else if (node.kind === ts.SyntaxKind.UndefinedKeyword) {
        return createRuntimePropertyRef('UndefinedKeyword');
    }
    else if (node.kind === ts.SyntaxKind.AnyKeyword) {
        return createRuntimePropertyRef('AnyKeyword');
    }
    else if (ts.isTypeOperatorNode(node) &&
        node.operator === ts.SyntaxKind.KeyOfKeyword) {
        return createRuntimeFunctionCall('KeyOfKeyword', [
            createMantaStyleRuntimeObject(node.type, typeParameters),
        ]);
    }
    else if (ts.isTypeReferenceNode(node)) {
        // Special Cases
        if (node.typeName.getText() === 'Array' &&
            node.typeArguments &&
            node.typeArguments.length === 1) {
            return createArrayType(ts.createArrayTypeNode(node.typeArguments[0]), typeParameters);
        }
        return createTypeReference(node, typeParameters);
    }
    else {
        throw new Error("Unhandled Type. " + node.getText());
    }
}
//# sourceMappingURL=utils.js.map